import modalWorker from './loadGltf.worker?worker';
import Dexie from 'dexie';
import { useModelStore } from '/@/store/modules/threejs';

export function initModalWorker() {
  type model = {
    modelName: string;
    modelVal: any;
  };

  const modalUrlArr = ['fm/fm-5.glb', 'fc/fc.glb'];

  const db: any = new Dexie('DB');
  window['CustomDB'] = db;

  const modelStore = useModelStore();
  db.version(1).stores({
    modal: '++id, modalName, modalVal',
  });

  new Promise(async (resolve) => {
    const validateModelUrlArr: string[] = [];
    for (let i = 0; i < modalUrlArr.length; i++) {
      const url = modalUrlArr[i];
      const modalName = url.replace(/(.*\/)*([^.]+).*/gi, '$2');
      const modalArr = await db.modal.where('modalName').equals(modalName).toArray();
      if (modalArr.length < 1) {
        console.log('utl------>', url);
        validateModelUrlArr.push(url);
      } else {
        if (!modelStore.modelArr.get(modalName)) {
          const model: model = {
            modelName: modalName,
            modelVal: modalArr[0].modalVal,
          };
          modelStore.setModel(model);
        }
      }
    }
    resolve(validateModelUrlArr);
  }).then((validateModelUrlArr: string[]) => {
    validateModelUrlArr.forEach((url) => {
      const worker = new modalWorker();

      worker.onmessage = async function (e) {
        const { data, message } = e.data;
        if (message === 'end') {
          setTimeout(() => {
            if (data) {
              modelStore.setModel(data);
            }
            worker.terminate();
          }, 1000);
        }
      };
      worker.postMessage({ message: 'load', data: url });
    });
  });
}

initModalWorker();
