import * as THREE from 'three';
import { TransformControls } from 'three/examples/jsm/controls/TransformControls'; // 引入模块
import { EffectComposer } from 'three/examples/jsm/postprocessing/EffectComposer.js';
import { RenderPass } from 'three/examples/jsm/postprocessing/RenderPass.js';
import { OutlinePass } from 'three/examples/jsm/postprocessing/OutlinePass.js';
import { FXAAShader } from 'three/examples/jsm/shaders/FXAAShader.js';
import { ShaderPass } from 'three/examples/jsm/postprocessing/ShaderPass.js';
import gsap from 'gsap';
// import * as dat from "dat.gui";

let modal,
  cubeList: THREE.Mesh[] = [],
  curve,
  line;

// //创建gui对象
// const gui = new dat.GUI();
// gui.domElement.style.top = '300'
// gui.domElement.style.clientTop = 300
// gui.domElement.style.zIndex = 9999

// 画点
const addCube = (initialPoints, scene) => {
  return initialPoints.map((pos) => {
    const geometry = new THREE.BoxBufferGeometry(1, 1, 1);
    const material = new THREE.MeshBasicMaterial({
      color: 0xffffff,
      side: THREE.DoubleSide,
    });
    const cube = new THREE.Mesh(geometry, material);
    cube.position.copy(pos);
    scene.add(cube);
    cubeList.push(cube);
    // gui.add(cube.position, 'x', -1000, 1000)
    // gui.add(cube.position, 'y', -1000, 1000)
    // gui.add(cube.position, 'z', -1000, 1000)
  });
};

// 获取点击位置
const addChangeEvent = () => {
  // debugger
  const control = new TransformControls(modal.camera, modal.renderer.domElement);
  // 获取点击位置
  const mouse = new THREE.Vector2();
  window.addEventListener(
    'mousedown',
    (event) => {
      // debugger
      mouse.x = (event.clientX / window.innerWidth) * 2 - 1;
      mouse.y = -(event.clientY / window.innerHeight) * 2 + 1;
      // mouse.x = ((event.clientX - dom.offsetLeft) / dom.clientWidth) * 2 - 1; // dom.offsetLeft -- dom元素距离浏览器左侧的距离   dom.clientWidth -- dom元素宽度
      // mouse.y = -((event.clientY - dom.offsetTop) / dom.clientHeight) * 2 + 1; // dom.offsetTop -- dom元素距离浏览器顶部的距离    dom.clientHeight -- dom元素高度
      //根据照相机，把这个向量转换到视点坐标系
      const vector = new THREE.Vector3(mouse.x, mouse.y, 0.5).unproject(modal.camera);

      // 方块点击检测
      // const rayCaster = new THREE.Raycaster();
      //在视点坐标系中形成射线,射线的起点向量是照相机， 射线的方向向量是照相机到点击的点，这个向量应该归一标准化。
      const rayCaster = new THREE.Raycaster(modal.camera.position, vector.sub(modal.camera.position).normalize());
      // rayCaster.setFromCamera(mouse, modal.camera);
      const intersects = rayCaster.intersectObjects(cubeList);
      if (intersects.length) {
        const target = intersects[0].object;
        control.attach(target); // 绑定controls和方块
        modal.scene.add(control);
      }
    },
    false
  );

  // 修改曲线后同步修改实体线条
  control.addEventListener('dragging-changed', (event) => {
    if (!event.value) {
      console.log(event.value);

      const points = curve.getPoints(50);
      line.geometry.setFromPoints(points);
    }
  });
};

// 画线
export const drawLine = (initialPoints, obj) => {
  modal = obj;
  addCube(initialPoints, modal.scene);
  curve = new THREE.CatmullRomCurve3(
    cubeList.map((cube) => cube.position) // 直接绑定方块的position以便后续用方块调整曲线
  );
  curve.curveType = 'chordal'; // 曲线类型
  curve.closed = false; // 曲线是否闭合

  const points = curve.getPoints(50); // 50等分获取曲线点数组
  line = new THREE.LineLoop(new THREE.BufferGeometry().setFromPoints(points), new THREE.LineBasicMaterial({ color: 0xffffff, linewidth: 10 })); // 绘制实体线条，仅用于示意曲线，后面的向量线条同理，相关代码就省略了

  modal.scene.add(line);

  addChangeEvent();
};

/* 设置模型居中 */
export const setModalCenter = (group) => {
  const box3 = new THREE.Box3();
  // 计算层级模型group的包围盒
  // 模型group是加载一个三维模型返回的对象，包含多个网格模型
  box3.expandByObject(group);
  // 计算一个层级模型对应包围盒的几何体中心在世界坐标中的位置
  const center = new THREE.Vector3();
  box3.getCenter(center);
  // console.log('查看几何体中心坐标', center);
  // 重新设置模型的位置，使之居中。
  group.position.x = group.position.x - center.x;
  group.position.y = group.position.y - center.y;
  group.position.z = group.position.z - center.z;
};

// 获取一个canvas 图文纹理
export const getTextCanvas = (w, h, textArr, imgUrl) => {
  // canvas 宽高最好是2的倍数
  const width = w;
  const height = h;
  // 创建一个canvas元素 获取上下文环境
  const canvas = document.createElement('canvas');
  canvas.style.letterSpacing = 10 + 'px';
  const ctx = canvas.getContext('2d') as CanvasRenderingContext2D;
  canvas.width = width;
  canvas.height = height;
  // 设置样式
  ctx.textAlign = 'start';
  // ctx.fillStyle = 'rgba(44, 62, 80, 0.65)';
  ctx.fillStyle = 'rgba(0, 0, 0, 0)';
  // 创建渐变
  // var gradient=ctx.createLinearGradient(0,0, canvas.width,0);
  // gradient.addColorStop(0,"magenta");
  // gradient.addColorStop(0.5,"blue");
  // gradient.addColorStop(1.0,"red");
  // // 用渐变填色
  // ctx.fillStyle=gradient;

  ctx.shadowColor = 'rgba(0, 10,0,0.8)';
  ctx.shadowBlur = 4;
  ctx.shadowOffsetX = 1;
  ctx.shadowOffsetY = 1;

  ctx.fillRect(0, 0, width, height);
  //添加背景图片，进行异步,否则可能会过早渲染,导致空白
  return new Promise((resolve, reject) => {
    if (imgUrl) {
      const img = new Image();

      img.src = new URL('../../assets/images/vent/model_image/' + imgUrl, import.meta.url).href;
      img.onload = () => {
        //将画布处理为透明
        ctx.clearRect(0, 0, width, height);
        //绘画图片
        ctx.drawImage(img, 0, 0, width, height);
        ctx.textBaseline = 'middle';
        // 由于文字需要自己排版 所以循环一下
        // item 是自定义的文字对象 包含文字内容 字体大小颜色 位置信息等
        textArr.forEach((item) => {
          ctx.font = item.font;
          ctx.fillStyle = item.color;
          ctx.fillText(item.text, item.x, item.y, 1024);
        });
        resolve(canvas);
      };
      //图片加载失败的方法
      img.onerror = (e) => {
        reject(e);
      };
    } else {
      //将画布处理为透明
      ctx.clearRect(0, 0, width, height);
      ctx.textBaseline = 'middle';

      textArr.forEach((item) => {
        ctx.lineWidth = 2;
        ctx.font = item.font;
        ctx.fillStyle = item.color;
        !!item.strokeStyle && (ctx.strokeStyle = item.strokeStyle);
        ctx.strokeText(item.text, item.x, item.y, 1024);
        ctx.fillText(item.text, item.x, item.y, 1024);
      });
      resolve(canvas);
    }
  });
};

// 发光路径
export const setLineGeo = (scene) => {
  const box = new THREE.BoxGeometry(30, 30, 30);
  // 立方体几何体box作为EdgesGeometry参数创建一个新的几何体
  const edges = new THREE.EdgesGeometry(box);
  // 立方体线框，不显示中间的斜线
  new THREE.TextureLoader().setPath('/model/hdr/').load('8.png', (texture) => {
    const edgesMaterial = new THREE.MeshBasicMaterial({
      // color: 0x00ffff,
      map: texture,
      transparent: true,
      depthWrite: false,
    });

    const line = new THREE.LineSegments(edges, edgesMaterial);
    // 网格模型和网格模型对应的轮廓线框插入到场景中
    scene.add(line);
  });

  debugger;
  box.attributes.position.array;

  const lightMaterial = new THREE.ShaderMaterial({
    vertexShader: `varying vec3 vPosition;
    varying vec2 vUv;
    uniform float uTime;
    
    void main(){
        // vec3 scalePosition = vec3(position.x+uTime,position.y,position.z+uTime);
        vec4 viewPosition = viewMatrix * modelMatrix * vec4(position,1);
        gl_Position = projectionMatrix *  viewPosition;
        vPosition = position;
        vUv = uv;
        
    }`,
    fragmentShader: `varying vec3 vPosition;
    varying vec2 vUv;
    uniform vec3 uColor;
    uniform float uHeight;
    vec4 permute(vec4 x)
    {
        return mod(((x*34.0)+1.0)*x, 289.0);
    }

    vec2 fade(vec2 t)
    {
        return t*t*t*(t*(t*6.0-15.0)+10.0);
    }
    float cnoise(vec2 P)
    {
        vec4 Pi = floor(P.xyxy) + vec4(0.0, 0.0, 1.0, 1.0);
        vec4 Pf = fract(P.xyxy) - vec4(0.0, 0.0, 1.0, 1.0);
        Pi = mod(Pi, 289.0); // To avoid truncation effects in permutation
        vec4 ix = Pi.xzxz;
        vec4 iy = Pi.yyww;
        vec4 fx = Pf.xzxz;
        vec4 fy = Pf.yyww;
        vec4 i = permute(permute(ix) + iy);
        vec4 gx = 2.0 * fract(i * 0.0243902439) - 1.0; // 1/41 = 0.024...
        vec4 gy = abs(gx) - 0.5;
        vec4 tx = floor(gx + 0.5);
        gx = gx - tx;
        vec2 g00 = vec2(gx.x,gy.x);
        vec2 g10 = vec2(gx.y,gy.y);
        vec2 g01 = vec2(gx.z,gy.z);
        vec2 g11 = vec2(gx.w,gy.w);
        vec4 norm = 1.79284291400159 - 0.85373472095314 * vec4(dot(g00, g00), dot(g01, g01), dot(g10, g10), dot(g11, g11));
        g00 *= norm.x;
        g01 *= norm.y;
        g10 *= norm.z;
        g11 *= norm.w;
        float n00 = dot(g00, vec2(fx.x, fy.x));
        float n10 = dot(g10, vec2(fx.y, fy.y));
        float n01 = dot(g01, vec2(fx.z, fy.z));
        float n11 = dot(g11, vec2(fx.w, fy.w));
        vec2 fade_xy = fade(Pf.xy);
        vec2 n_x = mix(vec2(n00, n01), vec2(n10, n11), fade_xy.x);
        float n_xy = mix(n_x.x, n_x.y, fade_xy.y);
        return 2.3 * n_xy;
    }
    
    void main(){
    
       float strength = (vPosition.y+uHeight/2.0)/uHeight;
        gl_FragColor = vec4(uColor,1.0 - strength);
      // float strength =1.0 - abs(cnoise(vUv * 10.0)) ;
      // gl_FragColor =vec4(strength,strength,strength,1);
    }`,
    transparent: true,
    side: THREE.DoubleSide,
  });
  const lightMesh = new THREE.Mesh(box, lightMaterial);
  lightMesh.geometry.computeBoundingBox();
  const { min, max } = lightMesh.geometry.boundingBox;
  const uHeight = max.y - min.y;
  lightMaterial.uniforms.uHeight = {
    value: uHeight,
  };

  lightMaterial.uniforms.uColor = {
    value: new THREE.Color(0x00ff00),
  };

  lightMaterial.uniforms.uTime = {
    value: 0,
  };
  // gsap.to(lightMesh.scale, {
  //   // x: 2,
  //   // z: 2,
  //   y: 0.8,
  //   duration: 1,
  //   ease: 'none',
  //   repeat: -1,
  //   yoyo: true,
  // });
  scene.add(lightMesh);
};

export const setOutline = (model, group) => {
  const { scene, renderer, camera } = model;
  const params = {
    edgeStrength: 10.0,
    edgeGlow: 1,
    edgeThickness: 1.0,
    pulsePeriod: 5,
    rotate: false,
    usePatternTexture: false,
  };

  const composer = new EffectComposer(renderer);

  const renderPass = new RenderPass(scene, camera);
  composer.addPass(renderPass);

  const outlinePass = new OutlinePass(new THREE.Vector2(window.innerWidth, window.innerHeight), scene, camera);
  composer.addPass(outlinePass);

  outlinePass.visibleEdgeColor.set(parseInt(0xffffff));
  outlinePass.hiddenEdgeColor.set('#190a05');
  outlinePass.edgeStrength = params.edgeStrength;
  outlinePass.edgeThickness = params.edgeThickness;
  outlinePass.pulsePeriod = params.pulsePeriod;
  outlinePass.usePatternTexture = params.usePatternTexture;

  // const textureLoader = new THREE.TextureLoader();
  // textureLoader.load('model/hdr/tri_pattern.jpg', function (texture) {
  //   outlinePass.patternTexture = texture;
  //   texture.wrapS = THREE.RepeatWrapping;
  //   texture.wrapT = THREE.RepeatWrapping;
  // });

  const effectFXAA = new ShaderPass(FXAAShader);
  effectFXAA.uniforms['resolution'].value.set(1 / window.innerWidth, 1 / window.innerHeight);
  composer.addPass(effectFXAA);

  const scale = 1;
  group.traverse(function (child) {
    if (child instanceof THREE.Mesh) {
      // child.geometry.center();

      child.geometry.computeBoundingSphere();
    }
  });

  // group.scale.multiplyScalar(Math.random() * 0.3 + 0.1);
  group.scale.divideScalar(scale);
  return { outlinePass, composer };
};
